package jp.mehve.nicofilter.filter.nicofilter;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;

import jp.mehve.httpproxy.ProxyHandlerConstants;
import jp.mehve.nicofilter.mainui.ConfigHolder;
import jp.mehve.nicofilter.mainui.InformationPanel;
import jp.mehve.io.PipeThread;

import compat.com.sun.net.httpserver.Filter;
import compat.com.sun.net.httpserver.Headers;
import compat.com.sun.net.httpserver.HttpExchange;

class NicoFilterException extends Exception {
    public NicoFilterException() {
        super();
    }

    public NicoFilterException(String message) {
        super(message);
    }

    public NicoFilterException(Exception cause) {
        super(cause);
    }

    public NicoFilterException(String message, Exception cause) {
        super(message, cause);
    }
}

/**
 * 
 * @author shinyak
 * 
 */
public class NicommentFilter extends Filter {
    private static final Pattern nicovideoAPIURLPattern = 
        Pattern.compile("http://msg[0-9]*\\.nicovideo\\.jp(:80)?/api/");
    private static final Pattern nicovideoFLVURLPattern =
        Pattern.compile("http://smile-\\w*[0-9]*.nicovideo.jp/smile\\?.*");
    
    
    private InformationPanel     panel                  = null;
    private List<String>         stolenCookies;

    private int                  filterRating;
    private Reputation           reputation;

    private Object               mutex                  = new Object();

    private ConfigHolder         config;
    private String               stolenThreadNo;

    public NicommentFilter(ConfigHolder config) {
        this.config = config;
        this.reputation = new Reputation(this);
        this.filterRating = -1;
    }

    public void clearReputation() {
        synchronized (config.mutex) {
            this.reputation.clear();
        }
    }

    public String getStolenCookie() {
        if (stolenCookies == null || stolenCookies.size() < 1) { return null; }
        return stolenCookies.get(0);
    }

    public String getStolenThreadNo() {
        return this.stolenThreadNo;
    }

    @Override
    public String description() {
        return "Modify Nicomment";
    }

    @Override
    public void doFilter(HttpExchange exchange, Chain chain) throws IOException {
        String uri = exchange.getRequestURI().toString();

        if (nicovideoAPIURLPattern.matcher(uri).matches()) {
            System.out.println("Matched Messaging API");

            PipeThread pt = new PipeThread(exchange.getResponseBody()) {
                @Override
                protected void pipe(InputStream is, OutputStream os) throws Exception {
                    filter(is, os);
                    byte[] buf = new byte[65536];
                    int len = 0;
                    while ((len = is.read(buf)) != 0) {
                        os.write(buf, 0, len);
                    }
                    os.flush();
                    os.close();
                }
            };

            exchange.getResponseHeaders().set(ProxyHandlerConstants.ATTR_FORCED_CHUNKED, "TRUE");

            PipedOutputStream pos = pt.getPipedOutputStream();
            exchange.setStreams(null, pos);
            pt.start();
            doSteal(exchange);

            if (panel != null) {
                panel.repaint();
            }
            chain.doFilter(exchange);
            pos.close();
        } else if (config.autosaveFLV && nicovideoFLVURLPattern.matcher(uri).matches()) {
            exchange.getResponseHeaders().set(ProxyHandlerConstants.ATTR_AUTOSAVE, "TRUE");
            chain.doFilter(exchange);
        } else {
            chain.doFilter(exchange);
        }
    }

    /**
     * steal cookie, url, and so on.
     * 
     * @param t
     */
    private synchronized void doSteal(HttpExchange t) {
        Headers headers = t.getRequestHeaders();

        // this.stolenURL = t.getRequestURI().toString();
        if (headers.containsKey("Cookie")) {
            this.stolenCookies = headers.get("Cookie");
            for (String val : this.stolenCookies) {
                System.err.println("STOLEN: " + val);
            }
        }
    }

    public void setPanel(InformationPanel panel) {
        this.panel = panel;
    }

    /**
     * NOTE: must be reentrant.
     * 
     * @param inputStream
     * @param outputStream
     * @throws NicoFilterException
     */
    public void filter(InputStream inputStream, OutputStream outputStream) throws NicoFilterException {
        try {
            SAXParser saxParser = SAXParserFactory.newInstance().newSAXParser();
            NicommentSAXHandler saxHandler = new NicommentSAXHandler();
            saxParser.parse(inputStream, saxHandler);

            Nicomment nicomment = saxHandler.getNicomment();

            // steal thread number here
            NicoThread th = nicomment.getNicoThread();
            if (th.thread != null) {
                synchronized (mutex) {
                    stolenThreadNo = th.thread;
                }
            }

            process(nicomment);

            outputStream.write(nicomment.toXMLStringWithXMLDecl().getBytes("UTF-8"));
            outputStream.flush();
        } catch (IOException e) {
            System.err.println("Comment Error: IOException");
            System.err.flush();
            throw new NicoFilterException(e);
        } catch (ParserConfigurationException e) {
            System.err.println("Comment Error: ParserConfigurationException");
            System.err.flush();
            throw new NicoFilterException(e);
        } catch (SAXException e) {
            System.err.println("Comment Error: SAXException");
            System.err.flush();
            e.printStackTrace();
        }
    }

    private void process(Nicomment nicomment) {
        ArrayList<NicoChat> chats = nicomment.getNicoChats();
        ArrayList<NicoChat> newChats = new ArrayList<NicoChat>();

        for (NicoChat chat : chats) {
            NicoChat newChat = chat.clone();
            try {
                synchronized (config.mutex) {
                    if (config.removeNGWord) {
                        for (String ng : config.ngWords) {
                            if (chat.comment.contains(ng)) {
                                chat.removed = true;
                                break;
                            }
                        }
                    }

                    if (config.removeNGID) {
                        if (config.ngIDs.contains(chat.userId)) {
                            chat.removed = true;
                        }
                    }

                    if (config.reputation) {
                        int userId = Integer.parseInt(chat.userId);
                        if (reputation.getRating(userId) <= filterRating) {
                            chat.removed = true;
                        }
                    }

                }

                if (panel != null) {
                    panel.addComment(chat);
                }
                if (!chat.removed) {
                    newChats.add(newChat);
                }
            } catch (NumberFormatException e) {
                // Remove a comment which throws this exception.
                // This is likely to be a code bug.
            }
        }

        nicomment.setNicoChats(newChats);
    }

}
